#

debug_output = False
rosDebug = False


import numpy as np
import os, sys
try:
    import gwtools
    import gwsurrogate as gws
    print "  gwsurrogate: ",  gws.__file__
    print "  gwtools: ",  gwtools.__file__
except:
    print " - no gwsurrogate - (almost everything from ROMWaveformManager will hard fail if you use it) "
try:
    import NRSur7dq2
    print "  NRSur7dq2: ", NRSur7dq2.__version__, NRSur7dq2.__file__
except:
    print " - no NRSur7dq2 - "

import lalsimulation as lalsim
import lal

from .. import lalsimutils
try:
    import LALHybrid
except:
    print " - no hybridization - "

from scipy.interpolate import interp1d
from scipy.linalg import inv
from scipy.interpolate import splrep as _splrep

import pickle
import h5py

try:
    dirBaseFiles =os.environ["GW_SURROGATE"] # surrogate base directory
except:
    print " ==> WARNING:  GW_SURROGATE environment variable is not set <== "
    print "    Only surrogates with direct implementation are available (NRSur7dq2) "
#    sys.exit(0)

print " ROMWaveformManager: ILE version"

#default_interpolation_kind = 'quadratic'  # spline interpolation   # very slow! 
default_interpolation_kind = 'linear'  # robust, fast

internal_ParametersAvailable ={}
# For each interesting simulation, store the definitions in a file
# Use 'execfile' to load those defintions now

MsunInSec = lal.MSUN_SI*lal.G_SI/lal.C_SI**3



#execfile(dirBaseFiles + "/"+"Sequence-GT-Aligned-UnequalMass/interface.py")

def myzero(arg):
    return 0

def RangeWrap1d(bound, val,fn):
     """
     RangeWrap1d: Uses np.piecewise to construct a piecewise function which is =fn inside the boundary, and 0 outside.
     SHOULD be syntactic sugar, but depending on the python version the language needed to implement this changes.
     """
# #    return (lambda x: fn(x) if  (x>bound[0] and x<bound[1]) else val)
# #  WARNING: piecewise is much faster, but will fail for numpy versions less than 1.8-ish :http://stackoverflow.com/questions/20800324/scipy-pchipinterpolator-error-array-cannot-be-safely-cast-to-required-type
# #     Unfortunately that is the version LIGO uses on their clusters.
     return (lambda x: np.piecewise( x,        [
                 np.logical_and(x> bound[0], x<bound[1]), 
                 np.logical_not(np.logical_and(x> bound[0], x<bound[1])) 
                 ], [fn, myzero]))
#    return (lambda x: np.where( np.logical_and(x> bound[0], x<bound[1]), fn(x),0))  # vectorized , but does not protect the call


def ModeToString(pair):
    return "l"+str(pair[0])+"_m"+str(pair[1]) 

def CreateCompatibleComplexOverlap(hlmf,**kwargs):
    modes = hlmf.keys()
    hbase = hlmf[modes[0]]
    deltaF = hbase.deltaF
    fNyq = np.max(lalsimutils.evaluate_fvals(hbase))
    if debug_output:
#        for key, value in kwargs.iteritems():
#            print (key, value)
        print kwargs
        print "dF, fNyq, npts = ",deltaF, fNyq, len(hbase.data.data)
    IP = lalsimutils.ComplexOverlap(fNyq=fNyq, deltaF=deltaF, **kwargs)
    return IP

def CreateCompatibleComplexIP(hlmf,**kwargs):
    """
    Creates complex IP (no maximization)
    """
    modes = hlmf.keys()
    hbase = hlmf[modes[0]]
    deltaF = hbase.deltaF
    fNyq = np.max(lalsimutils.evaluate_fvals(hbase))
    if debug_output:
#        for key, value in kwargs.iteritems():
#            print (key, value)
        print kwargs
        print "dF, fNyq, npts = ",deltaF, fNyq, len(hbase.data.data)
    IP = lalsimutils.ComplexIP(fNyq=fNyq, deltaF=deltaF, **kwargs)
    return IP

class NRError(Exception):
    """Base class for this module"""
    pass
class NRNoSimulation(NRError):
    """Nothing"""
    def __init__(self,expr,msg):
        print "No known simulation ", expr, msg
    pass

def SurrogateDimensionlessBasisFunction(sur,k):    
    def w(t):
        return sur.amp_fit_func(k,t)*np.exp(1j*sur.phase_fit_func(k,t))

    return w

def sur_identity(t,hp,hc):
    return t, hp, hc
def sur_conj(t,hp,hc):
    return t, hp, -hc


def ConvertWPtoSurrogateParams(P,**kwargs):
    """
    Takes P, returns arguments of the form usually used in gwsurrogate.
    (currently, just returns 1/q = P.m1/P.m1, the mass ratio parameter usually accepted)
    """
    q = P.m2/P.m1
#    return {"q":1./q}
    return 1./q

def ConvertWPtoSurrogateParamsAligned(P,**kwargs):
    """
    Takes P, returns arguments of the form used in gwsurrogate for a nonprecessing binary
    """
    q = P.m2/P.m1
    chi1 = np.array([0.0,0.0,P.s1z])
    chi2 = np.array([0.0,0.0,P.s2z])
    mtot=P.m1+P.m2
    tidal = {'Lambda1': P.lambda1,'Lambda2': P.lambda2}
    dist_mpc = P.dist/1e6/lal.PC_SI
    val =[1./q, chi1, chi2, mtot, dist_mpc, tidal]
    return val



def ConvertWPtoSurrogateParamsPrecessing(P,**kwargs):
    """
    Takes P, returns arguments of the form usually used in gwsurrogate.
    (currently, just returns 1/q = P.m1/P.m1, the mass ratio parameter usually accepted)
    """

    q = P.m2/P.m1
    chi1 = P.extract_param('chi1')
    theta1=phi1 =0
    if np.abs(chi1)>1e-5:
        theta1 = np.arccos( P.s1z/chi1)
        phi1 = np.arctan2(P.s1x,P.s1y)
#    return {"q":1./q, "chi1": chi1,"theta1":theta1,"phi1":phi1,"chi2z":P.s2z}
    val =np.array([1./q, chi1,theta1,phi1,P.s2z])
    return val

def ConvertWPtoSurrogateParamsPrecessingFull(P,**kwargs):
    """
    Takes P, returns arguments of the form usually used in gwsurrogate.
    (currently, just returns 1/q = P.m1/P.m1, the mass ratio parameter usually accepted)
    """

    q = P.m2/P.m1
    val =[1./q, np.array([P.s1x,P.s1y,P.s1z]), np.array([P.s2x,P.s2y,P.s2z]) ]
    return val



class WaveformModeCatalog:
    """
    Class containing ROM model.
    API is currently **unsafe** for precessing binaries (=ambiguous reference time)
    Reference for underlying notation:   Eq. (30) in http://arxiv.org/pdf/1308.3565v2
       group
       param
       lmax                          # specifies modes to attempt to load. Not guaranteed to/required to find all.
       strain_basis_functions_dimensionless   # don't recall
       mode_list_to_load        # ability to constrain the mode list.  Passed directly to low-level code
       build_fourier_time_window  # window for FT. NOT USED
       reflection_symmetric     # reflection symmetry used
       max_nbasis_per_mode   # constrain basis size
       coord_names_internal    # coordinate names used by the basis.  FUTURE

    """


    def __init__(self, group ,param, lmax=2, 
                 strain_basis_functions_dimensionless=None,
                 mode_list_to_load=None,build_fourier_time_window=1000,reflection_symmetric=True,max_nbasis_per_mode=None,coord_names_internal=['q']):
        self.group  = group
        self.param = param 
        self.deltaToverM =0
        self.lmax =lmax
        self.coord_names=coord_names_internal
        self.fOrbitLower =0.    #  Used to clean results.  Based on the phase of the 22 mode
        self.fMinMode ={}
        self.sur_dict = {}
        self.post_dict = {}
        self.post_dict_complex ={}
        self.post_dict_complex_coef ={}
        self.parameter_convert = {}
        self.single_mode_sur = True
        self.nbasis_per_mode ={}   # number of basis functions
        self.reflection_symmetric = reflection_symmetric

        lm_list=None
        lm_list = []
        if rosDebug:
            print " WARNING: Using a restricted mode set requires a custom modification to gwsurrogate "
        Lmax =lmax
        for l in np.arange(2,Lmax+1):
            for m in np.arange(-l,l+1):
                if m<0 and reflection_symmetric:
                    continue
                lm_list.append( (l,m))
        if not(mode_list_to_load is None):
            lm_list = mode_list_to_load  # overrride
        if rosDebug:
            print " ROMWaveformManager: Loading restricted mode set ", lm_list

        my_converter = ConvertWPtoSurrogateParams
        if 'NRSur4d' in param:
            print " GENERATING ROM WAVEFORM WITH SPIN PARAMETERS "
            my_converter = ConvertWPtoSurrogateParamsPrecessing
            reflection_symmetric=False
        if 'NRHyb' in param and not'Tidal' in param:
            print " GENERATING hybrid ROM WAVEFORM WITH ALIGNED SPIN PARAMETERS "
            my_converter = ConvertWPtoSurrogateParamsAligned
            self.single_mode_sur=False
        if 'Tidal' in param:
            print " GENERATING hybrid ROM WAVEFORM WITH ALIGNED SPIN AND TIDAL PARAMETERS "
            my_converter = ConvertWPtoSurrogateParamsAligned
            self.single_mode_sur=False
        if 'NRSur7d' in param:
            if  rosDebug:
                print " GENERATING ROM WAVEFORM WITH FULL SPIN PARAMETERS "
            my_converter = ConvertWPtoSurrogateParamsPrecessingFull
            self.single_mode_sur=False
            reflection_symmetric=False
        # PENDING: General-purpose interface, based on the coordinate string specified. SHOULD look up these names from the surrogate!
        def convert_coords(P):
            vals_out = np.zeros(len(coord_names_internal))
            for indx in np.arange(len(coord_names_internal)):
                vals_out[indx] = P.extract_param( coord_names_internal[indx])
                if coord_names_internal[indx] == 'q':
                    vals_out[indx] = 1./vals_out[indx]
                return vals_out

        raw_modes =[]
        if self.single_mode_sur: #(not 'NRSur7d' in param) and (not 'NRHyb' in param):
            self.sur =  gws.EvaluateSurrogate(dirBaseFiles +'/'+group+param,use_orbital_plane_symmetry=reflection_symmetric, ell_m=None) # lm_list) # straight up filename.  MODIFY to change to use negative modes
                # Modified surrogate import call to load *all* modes all the time
            raw_modes = self.sur.all_model_modes()
            self.modes_available=[]
        elif 'NRHybSur' in param:
            if 'Tidal' in param:
                self.sur = gws.LoadSurrogate(dirBaseFiles +'/'+group,surrogate_name_spliced=param)   # get the dimensinoless surrogate file?
            else:
                self.sur = gws.LoadSurrogate(dirBaseFiles +'/'+group+param)   # get the dimensinoless surrogate file?
            raw_modes = self.sur._sur_dimless.mode_list  # raw modes
            reflection_symmetric = True
            self.modes_available=[]
#            self.modes_available=[(2, 0), (2, 1), (2,-1), (2, 2),(2,-2), (3, 0), (3, 1),(3,-1), (3, 2),(3,-2), (3, 3),(3,-3), (4, 2),(4,-2), (4, 3),(4,-3), (4, 4), (4,-4),(5, 5), (5,-5)]  # see sur.mode_list
            t = self.sur._sur_dimless.domain
            self.ToverMmin = t.min()
            self.ToverMmax = t.max()
            self.ToverM_peak=0   # Need to figure out where this is?  Let's assume it is zero to make my life easier
            # for mode in self.modes_available:
            #     # Not used, bt populate anyways
            #     self.post_dict[mode] = sur_identity
            #     self.post_dict_complex[mode]  = lambda x: x   # to mode
            #     self.post_dict_complex_coef[mode] = lambda x:x  #  to coefficients.
            #     self.parameter_convert[mode] =  my_converter #  ConvertWPtoSurrogateParams   # default conversion routine
#            return
        elif 'NRSur7dq4' in param:
            print param
            self.sur = gws.LoadSurrogate(dirBaseFiles +'/'+group+param)   # get the dimensinoless surrogate file?
            raw_modes = self.sur._sur_dimless.mode_list  # raw modes
            reflection_symmetric = False
            self.modes_available=[]
            print raw_modes
            self.modes_available=raw_modes
            t = self.sur._sur_dimless.t_coorb
            self.ToverMmin = t.min()
            self.ToverMmax = t.max()
            self.ToverM_peak=0   # Need to figure out where this is?  Let's assume it is zero to make my life easier
            for mode in raw_modes:
            #     # Not used, bt populate anyways
                self.post_dict[mode] = sur_identity
                self.post_dict_complex[mode]  = lambda x: x   # to mode
                self.post_dict_complex_coef[mode] = lambda x:x  #  to coefficients.
                self.parameter_convert[mode] =  my_converter #  ConvertWPtoSurrogateParams   # default conversion routine
            return
        else:
            self.sur = NRSur7dq2.NRSurrogate7dq2()
            reflection_symmetric = False
            self.modes_available = [(2, -2), (2, -1), (2, 0), (2, 1), (2, 2), (3, -3), (3, -2), (3, -1), (3, 0), (3, 1), (3, 2), (3, 3), (4, -4), (4, -3), (4, -2), (4, -1), (4, 0), (4, 1), (4, 2), (4, 3), (4, 4)];    
            t = self.sur.t_coorb
            self.ToverMmin = t.min()
            self.ToverMmax = t.max()
            self.ToverM_peak=0
            for mode in self.modes_available:
                # Not used, bt populate anyways
                self.post_dict[mode] = sur_identity
                self.post_dict_complex[mode]  = lambda x: x   # to mode
                self.post_dict_complex_coef[mode] = lambda x:x  #  to coefficients.
                self.parameter_convert[mode] =  my_converter #  ConvertWPtoSurrogateParams   # default conversion routine
            return
        # Load surrogates from a mode-by-mode basis, and their conjugates
        for mode in raw_modes:
          if mode[0]<=self.lmax and mode in lm_list:  # latter SHOULD be redundant (because of ell_m=lm_list)
            print " Loading mode ", mode
            self.modes_available.append(mode)
            self.post_dict[mode] = sur_identity
            self.post_dict_complex[mode]  = lambda x: x   # to mode
            self.post_dict_complex_coef[mode] = lambda x:x  #  to coefficients.
            self.parameter_convert[mode] =  my_converter #  ConvertWPtoSurrogateParams   # default conversion routine
            if self.single_mode_sur:
                self.sur_dict[mode] = self.sur.single_mode(mode)
                print ' mode ', mode, self.sur_dict[mode].B.shape
                self.nbasis_per_mode[mode] = (self.sur_dict[mode].B.shape)[1]
            if max_nbasis_per_mode != None  and self.sur_dict[mode].surrogate_mode_type == 'waveform_basis':
             if max_nbasis_per_mode >0: # and max_nbasis_per_mode < self.nbasis_per_mode[mode]:
                # See  https://arxiv.org/pdf/1308.3565v2.pdf  Eqs. 13 - 19
                # Must truncate *orthogonal* basis.
                # Works only for LINEAR basis
                # B are samples of the basis on some long time, V
                sur = self.sur_dict[mode]
                print " Truncating basis for mode ", mode, " to size ", max_nbasis_per_mode,  " but note the number of EIM points remains the same..."
                V = self.sur_dict[mode].V
                n_basis = len(V)
                V_inv = inv(V)
                mtx_E = np.dot(self.sur_dict[mode].B,V)
#                print "E ", mtx_E.shape
                # Zero out the components we don't want
                if max_nbasis_per_mode < n_basis:
                    mtx_E[:,max_nbasis_per_mode:n_basis] *=0
                # Regenerate
                sur.B = np.dot(mtx_E , V_inv)
                sur.reB_spline_params = [_splrep(sur.times, sur.B[:,jj].real, k=3) for jj in range(sur.B.shape[1])]
                sur.imB_spline_params = [_splrep(sur.times, sur.B[:,jj].imag, k=3) for jj in range(sur.B.shape[1])]
                self.nbasis_per_mode[mode] = len(self.sur_dict[mode].V) # if you truncate over the orthogonal basis, you still need to use the fit at all the EIM points!
                # This SHOULD update the copies inside the surrogate, so the later interpolation called by EvaluateSingleModeSurrogate will interpolate this data
                
            if reflection_symmetric and raw_modes.count((mode[0],-mode[1]))<1:
                mode_alt = (mode[0],-mode[1])
                if rosDebug:
                    print " Adjoining postprocessing to enable complex conjugate for reflection symmetric case", mode_alt
#                if max_nbasis_per_mode:
 #                   self.nbasis_per_mode[mode_alt] = np.max([int(max_nbasis_per_mode),1])    # INFRASTRUTCTURE PLAN: Truncate t                print " Loading mode ", mode_alt, " via reflection symmetry "
                self.modes_available.append(mode_alt)
                self.post_dict[mode_alt] = sur_conj
                self.post_dict_complex_coef[mode_alt] = lambda x,l=mode[0]: np.power(-1,l)*np.conj(x)  # beware, do not apply this twice.
                self.post_dict_complex[mode_alt] = np.conj  # beware, do not apply this twice.
                self.parameter_convert[mode_alt] = my_converter
                if self.single_mode_sur: 
                    self.nbasis_per_mode[mode_alt] = self.nbasis_per_mode[mode]
                    self.sur_dict[mode_alt] = self.sur_dict[mode]
        if not self.single_mode_sur:
            # return after performing all the neat reflection symmetrization setup described above, in case model is *not* a single-mode surrogate
            print "  ... done setting mode symmetry requirements", self.modes_available
#            print raw_modes, self.post_dict
            return  
        # CURRENTLY ONLY LOAD THE 22 MODE and generate the 2,-2 mode by symmetr
        t = self.sur_dict[(2,2)].times  # end time
        self.ToverMmin = t.min()
        self.ToverMmax = t.max()
        P=lalsimutils.ChooseWaveformParams() # default is q=1 object
        params_tmp = self.parameter_convert[(2,2)](P)
        if rosDebug:
            print " Passing temporary parameters ", params_tmp, " to find the peak time default "
        # print dir(self.sur_dict[(2,2)])
        # print self.sur_dict[(2,2)].__dict__.keys()
        # print self.sur_dict[(2,2)].parameterization

#        t, hp, hc = self.sur_dict[(2,2)]( **params_tmp  );   # calculate merger time -- addresses conventions on peak time location, and uses named arguments
        t, hp, hc = self.sur_dict[(2,2)]( params_tmp  );   # calculate merger time -- addresses conventions on peak time location, and uses named arguments
        self.ToverM_peak = t[np.argmax(np.abs(hp**2+hc**2))]  # discrete maximum time. Sanity check
        if rosDebug:
            print " Peak time for ROM ", self.ToverM_peak

        # BASIS MANAGEMENT: Not yet implemented
        # Assume a DISTINCT BASIS SET FOR AL MODES, which will be ANNOYING
        self.strain_basis_functions_dimensionless_data ={}
        self.strain_basis_functions_dimensionless = self.sur_dict[(2,2)].resample_B  # We may need to add complex conjugate functions too. And a master index for basis functions associated with different modes

    def print_params(self):
        print " Surrogate model "
        print "   Modes available "
        for mode in self.sur_dict:
            print  "    " , mode, " nbasis = ", self.nbasis_per_mode[mode]

    # same arguments as hlm
    def complex_hoft(self, P, force_T=False, deltaT=1./16384, time_over_M_zero=0.,sgn=-1):
        hlmT = self.hlmoft(P, force_T, deltaT,time_over_M_zero)
        npts = hlmT[(2,2)].data.length
        wfmTS = lal.CreateCOMPLEX16TimeSeries("h", lal.LIGOTimeGPS(0.), 0., deltaT, lalsimutils.lsu_DimensionlessUnit, npts)
        wfmTS.data.data[:] = 0   # SHOULD NOT BE NECESARY, but the creation operator doesn't robustly clean memory
        wfmTS.epoch = hlmT[(2,2)].epoch
        for mode in hlmT.keys():
            # PROBLEM: Be careful with interpretation. The incl and phiref terms are NOT tied to L.
            if rosDebug:
                print mode, np.max(hlmT[mode].data.data), " running max ",  np.max(np.abs(wfmTS.data.data))
            wfmTS.data.data += np.exp(-2*sgn*1j*P.psi)* hlmT[mode].data.data*lal.SpinWeightedSphericalHarmonic(P.incl,-P.phiref,-2, int(mode[0]),int(mode[1]))
        return wfmTS
    def complex_hoff(self,P, force_T=False):
        htC  = self.complex_hoft(P, force_T=force_T,deltaT= P.deltaT)
        TDlen = int(1./P.deltaF * 1./P.deltaT)
        assert TDlen == htC.data.length
        hf = lal.CreateCOMPLEX16FrequencySeries("Template h(f)",
                                                htC.epoch, htC.f0, 1./htC.deltaT/htC.data.length, lalsimutils.lsu_HertzUnit, 
                                                htC.data.length)
        fwdplan=lal.CreateForwardCOMPLEX16FFTPlan(htC.data.length,0)
        lal.COMPLEX16TimeFreqFFT(hf, htC, fwdplan)
        return hf
    def real_hoft(self,P,Fp=None, Fc=None):
        """
        Returns the real-valued h(t) that would be produced in a single instrument.
        Translates epoch as needed.
        Based on 'hoft' in lalsimutils.py
        """
        # Create complex timessereis
        htC = self.complex_hoft(P,force_T=1./P.deltaF, deltaT= P.deltaT)  # note P.tref is NOT used in the low-level code
        TDlen  = htC.data.length
        if rosDebug:
            print  "Size sanity check ", TDlen, 1/(P.deltaF*P.deltaT)
            print " Raw complex magnitude , ", np.max(htC.data.data)
            
        # Create working buffers to extract data from it -- wasteful.
        hp = lal.CreateREAL8TimeSeries("h(t)", htC.epoch, 0.,
            P.deltaT, lalsimutils.lsu_DimensionlessUnit, TDlen)
        hc = lal.CreateREAL8TimeSeries("h(t)", htC.epoch, 0.,
            P.deltaT, lalsimutils.lsu_DimensionlessUnit, TDlen)
        hT = lal.CreateREAL8TimeSeries("h(t)", htC.epoch, 0.,
            P.deltaT, lalsimutils.lsu_DimensionlessUnit, TDlen)
        # Copy data components over
        #  - note htC is hp - i hx
        hp.data.data = np.real(htC.data.data)
        hc.data.data = (-1) * np.imag(htC.data.data)
        # transform as in lalsimutils.hoft
        if Fp!=None and Fc!=None:
            hp.data.data *= Fp
            hc.data.data *= Fc
            hp = lal.AddREAL8TimeSeries(hp, hc)
            hoft = hp
        elif P.radec==False:
            fp = lalsimutils.Fplus(P.theta, P.phi, P.psi)
            fc = lalsimutils.Fcross(P.theta, P.phi, P.psi)
            hp.data.data *= fp
            hc.data.data *= fc
            hp.data.data  = lal.AddREAL8TimeSeries(hp, hc)
            hoft = hp
        else:
            # Note epoch must be applied FIRST, to make sure the correct event time is being used to construct the modulation functions
            hp.epoch = hp.epoch + P.tref
            hc.epoch = hc.epoch + P.tref
            if rosDebug:
                print " Real h(t) before detector weighting, ", np.max(hp.data.data), np.max(hc.data.data)
            hoft = lalsim.SimDetectorStrainREAL8TimeSeries(hp, hc,    # beware, this MAY alter the series length??
                P.phi, P.theta, P.psi, 
                lalsim.DetectorPrefixToLALDetector(str(P.detector)))
            hoft = lal.CutREAL8TimeSeries(hoft, 0, hp.data.length)       # force same length as before??
            if rosDebug:
                print "Size before and after detector weighting " , hp.data.length, hoft.data.length
        if rosDebug:
            print  " Real h_{IFO}(t) generated, pre-taper : max strain =", np.max(hoft.data.data)
        if P.taper != lalsimutils.lsu_TAPER_NONE: # Taper if requested
            lalsim.SimInspiralREAL8WaveTaper(hoft.data, P.taper)
        if P.deltaF is not None:
            TDlen = int(1./P.deltaF * 1./P.deltaT)
            print "Size sanity check 2 ", int(1./P.deltaF * 1./P.deltaT), hoft.data.length
            assert TDlen >= hoft.data.length
            npts = hoft.data.length
            hoft = lal.ResizeREAL8TimeSeries(hoft, 0, TDlen)
            # Zero out the last few data elements -- NOT always reliable for all architectures; SHOULD NOT BE NECESSARY
            hoft.data.data[npts:TDlen] = 0

        if rosDebug:
            print  " Real h_{IFO}(t) generated : max strain =", np.max(hoft.data.data)
        return hoft

    def non_herm_hoff(self,P):
        """
        Returns the 2-sided h(f) associated with the real-valued h(t) seen in a real instrument.
        Translates epoch as needed.
        Based on 'non_herm_hoff' in lalsimutils.py
        """
        htR = self.real_hoft() # Generate real-valued TD waveform, including detector response
        if P.deltaF == None: # h(t) was not zero-padded, so do it now
            TDlen = nextPow2(htR.data.length)
            htR = lal.ResizeREAL8TimeSeries(htR, 0, TDlen)
        else: # Check zero-padding was done to expected length
            TDlen = int(1./P.deltaF * 1./P.deltaT)
            assert TDlen == htR.data.length
        fwdplan=lal.CreateForwardCOMPLEX16FFTPlan(htR.data.length,0)
        htC = lal.CreateCOMPLEX16TimeSeries("hoft", htR.epoch, htR.f0,
            htR.deltaT, htR.sampleUnits, htR.data.length)
        # copy h(t) into a COMPLEX16 array which happens to be purely real
        htC.data.data[:htR.data.length] = htR.data.data
#        for i in range(htR.data.length):
#            htC.data.data[i] = htR.data.data[i]
        hf = lal.CreateCOMPLEX16FrequencySeries("Template h(f)",
            htR.epoch, htR.f0, 1./htR.deltaT/htR.data.length, lalsimutils.lsu_HertzUnit, 
            htR.data.length)
        lal.COMPLEX16TimeFreqFFT(hf, htC, fwdplan)
        return hf


    def estimateFminHz(self,P,fmin=10.):
        # This SHOULD use information from the ROM
        return 2*self.fMin/(MsunInSec*(P.m1+P.m2)/lal.MSUN_SI)

    def estimateDurationSec(self,P,fmin=10.):
        """
        estimateDuration uses fmin*M from the (2,2) mode to estimate the waveform duration from the *well-posed*
        part.  By default it uses the *entire* waveform duration.
        CURRENTLY DOES NOT IMPLEMENT frequency-dependent duration
        """
        return (self.ToverMmax - self.ToverMmin)*MsunInSec*(P.m1+P.m2)/lal.MSUN_SI
        return None

    def basis_oft(self,  P, force_T=False, deltaT=1./16384, time_over_M_zero=0.,return_numpy=False):
        m_total_s = MsunInSec*(P.m1+P.m2)/lal.MSUN_SI
        # Create a suitable set of time samples.  Zero pad to 2^n samples.
        T_estimated = np.abs(self.sur_dict[(2,2)].tmin)*m_total_s
        print " Estimated duration ", T_estimated
#        T_estimated =  20 # FIXME. Time in seconds
        npts=0
        if not force_T:
            npts_estimated = int(T_estimated/deltaT)
            npts = lalsimutils.nextPow2(npts_estimated)
        else:
            npts = int(force_T/deltaT)
            if rosDebug:
                print " Forcing length T=", force_T, " length ", npts
        tvals = (np.arange(npts)-npts/2)*deltaT   # Use CENTERED time to make sure I handle CENTERED NR signal (usual)
        if rosDebug:
            print " time range being sampled ", [min(tvals),max(tvals)], " corresponding to dimensionless range", [min(tvals)/m_total_s,max(tvals)/m_total_s], " for mtotal ", (P.m1+P.m2)/lal.MSUN_SI

        # Identify natural time region, and prepare to taper accordingly

        # Basis function matrix on grid
        # BEWARE: we need to zero out anything not well within the region of validity. 
        basis_grid = {}
        for mode in self.sur_dict:
            new_B = self.sur_dict[mode].resample_B(tvals/m_total_s + self.ToverM_peak,ext=1)  # should return 0 outside the domain

            # Identify start and end time, for tapering if desired (probably not)
            tmin = (self.sur_dict[mode].tmin-self.ToverM_peak)*m_total_s
            tmax = (self.sur_dict[mode].tmax-self.ToverM_peak)*m_total_s
            for indx in np.arange(new_B.shape[1]):
                if indx >= self.nbasis_per_mode[mode]:    # Allow truncation of model. Note fencepost
                    continue
                how_to_store = (mode[0], mode[1], indx)
                if rosDebug:
                    print " Storing resampled copy : ", how_to_store, " expected start and end ", tmin, tmax, " for mtot = ", (P.m1+P.m2)/lal.MSUN_SI
                basis_grid[how_to_store] = self.post_dict_complex[mode](new_B[:,indx])

        # Optional: return just this way
        if return_numpy:
            return tvals, basis_grid

        # Otherwise, pack into a set of LAL data structures, with epoch etc set appropriately, so we can use in ILE without changes
        ret = {}

        # Taper design.  Note taper needs to be applied IN THE RIGHT PLACE.  
        print " Tapering basis part 1: start "
        vectaper = np.zeros(len(tvals))
        t_start_taper = (self.ToverMmin - self.ToverM_peak)*m_total_s#-self.ToverM_peak*m_total_s
        t_end_taper = t_start_taper +100*m_total_s  # taper over around 100 M
        print "  t interval (s) =  ", t_start_taper, t_end_taper
#        print t_start_taper, t_end_taper, tvals
        def fn_taper(x):
            return np.piecewise(x,[x<t_start_taper,x>t_end_taper,np.logical_and(x>=t_start_taper, x<=t_end_taper)], [0,1,lambda z: 0.5-0.5*np.cos(np.pi* (z-t_start_taper)/(t_end_taper-t_start_taper))])
        vectaper= fn_taper(tvals)

        print " Tapering basis part 2: end "  # IMPORTANT: ROMs can be discontinuous at end, which casues problems
        vectaper2 = np.zeros(len(tvals))
        t_start_taper = (self.ToverMmax-self.ToverM_peak-10)*m_total_s  # taper over last 5 M
        t_end_taper = (self.ToverMmax-self.ToverM_peak)*m_total_s  
        print "  t interval (s) =  ", t_start_taper, t_end_taper
#        print t_start_taper, t_end_taper, tvals
        def fn_taper2(x):
            return np.piecewise(x,[x<t_start_taper,x>t_end_taper,np.logical_and(x>=t_start_taper, x<=t_end_taper)], [1,0,lambda z: 0.5+0.5*np.cos(np.pi* (z-t_start_taper)/(t_end_taper-t_start_taper))])
        vectaper2= fn_taper2(tvals)

        for ind in basis_grid:
            wfmTS = lal.CreateCOMPLEX16TimeSeries("h", lal.LIGOTimeGPS(0.), 0., deltaT, lalsimutils.lsu_DimensionlessUnit, npts)
            wfmTS.epoch = tvals[0]
            wfmTS.data.data = vectaper*basis_grid[ind]
            wfmTS.data.data *= vectaper2
            ret[ind] = wfmTS

        return ret

    def basis_off(self,  P, force_T=False, deltaT=1./16384, time_over_M_zero=0.):
        """
        basis_off takes fourier transforms of LAL timeseries generated from basis_oft.
        All modes have physical units, appropriate to a physical signal.
        """
        # Code is operationally identical to hlmoff, hence the names
        hlmF ={}
        hlmT = self.basis_oft(P,force_T=force_T,deltaT=deltaT,time_over_M_zero=time_over_M_zero, return_numpy=False)
        for mode in hlmT.keys():
            wfmTS=hlmT[mode]
            # Take the fourier transform
            wfmFD = lalsimutils.DataFourier(wfmTS)  # this creates a new f series for *each* call.
            # Store the resulting mode
            hlmF[mode] = wfmFD
        return hlmF


    def coefficients(self,P,**kwargs):
        """
        Returns the values of the ROM coefficients for the parameter P.  
        Usees the key-value pairing convention described in basis_oft.
        Note that the coefficients are created on a mode by mode basis, so the conversion call is VERY redundant.
        """

        coefs = {}
        for mode in self.sur_dict:
            params = self.parameter_convert[mode](P,**kwargs)  # very redundant, almost certainly the same for all modes in the surrogate.
            params_surrogate = self.sur_dict[mode].get_surr_params(params)
            if rosDebug:
                print " passing params to mode : ", mode, params
                print " surrogate natural parameter is ", params_surrogate

            # New version: gw-surrogate-0.5
            h_EIM = self.sur_dict[mode].eim_coeffs(params_surrogate, 'waveform_basis')
            # OLD VERSION: gw-surrogate-0.4.2 and earlier
            # x0= self.sur_dict[mode]._affine_mapper(params_surrogate)
            # amp_eval = self.sur_dict[mode]._amp_eval(x0)
            # phase_eval = self.sur_dict[mode]._phase_eval(x0)
            # norm_eval = self.sur_dict[mode]._norm_eval(x0)
            # h_EIM = np.zeros(len(amp_eval))
            # if self.sur_dict[mode].surrogate_mode_type  == 'waveform_basis':
            #     h_EIM = norm_eval*amp_eval*np.exp(1j*phase_eval)


            for indx in np.arange(self.nbasis_per_mode[mode]):  
                how_to_store = (mode[0], mode[1], indx)
                coefs[how_to_store]  = self.post_dict_complex_coef[mode](h_EIM[indx])   # conjugation as needed

        return coefs

    # See NR code 
    def hlmoft(self,  P, force_T=False, deltaT=1./16384, time_over_M_zero=0.,use_basis=False,Lmax=np.inf,hybrid_time=None,hybrid_use=False,hybrid_method='taper_add',hybrid_frequency=None,verbose=False,rom_taper_start=False,rom_taper_end=True,use_reference_spins=False):
        """
        hlmoft uses the dimensionless ROM basis functions to extract hlm(t) in physical units, in a LAL array.
        The argument 'P' is a ChooseWaveformParaams object
        FIXME: Add tapering option!

        rom_taper_start  # use manual tapering in this routine.  Should taper if this is true, OR the P.taper includes start tapering
        rom_taper_end  # use gwsurrogate built-in tapering as appropriate (HybSur only), with hard-coded taper time from scott (40M out of 140 M)
        """
        hybrid_time_viaf = hybrid_time
        hlmT ={}
        # Define units
        m_total_s = MsunInSec*(P.m1+P.m2)/lal.MSUN_SI
        distance_s = P.dist/lal.C_SI  # insures valid units

        # Create a suitable set of time samples.  Zero pad to 2^n samples.
        T_estimated =  20 # FIXME. Time in seconds
        npts=0
        if not force_T:
            npts_estimated = int(T_estimated/deltaT)
            npts = lalsimutils.nextPow2(npts_estimated)
        else:
            npts = int(force_T/deltaT)
            if rosDebug:
                print " Forcing length T=", force_T, " length ", npts
        tvals = (np.arange(npts)-npts/2)*deltaT   # Use CENTERED time to make sure I handle CENTERED NR signal (usual)
        if rosDebug:
            print " time range being sampled ", [min(tvals),max(tvals)], " corresponding to dimensionless range", [min(tvals)/m_total_s,max(tvals)/m_total_s]

        # Populate basis coefficients, if we need it
        if use_basis:
            bT = self.basis_oft(P, force_T=force_T, deltaT=deltaT, time_over_M_zero=time_over_M_zero)
            coefs = self.coefficients(P)


        # extract arguments
        q = P.m1/P.m2  # NOTE OPPOSITE CONVENTION ADOPTED BY SURROGATE GROUP. Code WILL FAIL if q<1
        if q<1:
            q=1./q  # this flips the objects, swapping phase, but it is a good place to start

        # Option 0: Use NRSur7dsq approach (i.e., generate an hlmoft dictionary)
        hlmT_dimensionless={}
        if 'NRSur7dq2' in self.param:
            params_here = self.parameter_convert[(2,2)](P)
            tvals_dimensionless= tvals/m_total_s + self.ToverM_peak
            indx_ok = np.logical_and(tvals_dimensionless  > self.ToverMmin , tvals_dimensionless < self.ToverMmax)
            hlmT ={}
            taper_end_duration =None
            if rom_taper_end:
                taper_end_duration =40.0
            if P.fref >0 and use_reference_spins:
                hlmT_dimensionless_narrow = self.sur(params_here[0], params_here[1],params_here[2],f_ref=P.fref, MTot=(P.m1+P.m2)/lal.MSUN_SI, t=tvals_dimensionless[indx_ok]*m_total_s) #,f_low=0)   
            else:
                hlmT_dimensionless_narrow = self.sur(params_here[0], params_here[1],params_here[2],t=tvals_dimensionless[indx_ok]) #,f_low=0)
            for mode in self.modes_available:
                hlmT_dimensionless[mode] = np.zeros(len(tvals_dimensionless),dtype=complex)
                hlmT_dimensionless[mode][indx_ok] = hlmT_dimensionless_narrow[mode]
           
        if 'NRSur7dq4' in self.param:
            print self.sur
            params_here = self.parameter_convert[(2,2)](P)
            tvals_dimensionless= tvals/m_total_s + self.ToverM_peak
            indx_ok = np.logical_and(tvals_dimensionless  > self.ToverMmin , tvals_dimensionless < self.ToverMmax)
            hlmT ={}
            taper_end_duration =None
            if rom_taper_end:
                taper_end_duration =40.0
                print params_here[0],params_here[1],params_here[2]
            if P.fref >0 and use_reference_spins:
                time,hlmT_dimensionless_narrow,dym = self.sur(params_here[0], params_here[1],params_here[2],f_ref=P.fref, MTot=(P.m1+P.m2)/lal.MSUN_SI, times=tvals_dimensionless[indx_ok]*m_total_s,f_low=0,taper_end_duration=taper_end_duration) #,f_low=0)
            else:
                time,hlmT_dimensionless_narrow,dym = self.sur(params_here[0],params_here[1],params_here[2],times=tvals_dimensionless[indx_ok],f_low=0,taper_end_duration=taper_end_duration)
            for mode in self.modes_available:
                hlmT_dimensionless[mode] = np.zeros(len(tvals_dimensionless),dtype=complex)
                hlmT_dimensionless[mode][indx_ok] = hlmT_dimensionless_narrow[mode]
        # Option 1: Use NRHybXXX approach (i.e., generate an hlmoft dictionary...but with its OWN time grid and scaling...very annoying)
        if 'NRHyb' in self.param:
            params_here = self.parameter_convert[(2,2)](P)
            f_low = P.fmin  # need to convert to dimensionless time
            tvals_dimensionless= tvals/m_total_s + self.ToverM_peak
            indx_ok = np.logical_and(tvals_dimensionless  > self.ToverMmin , tvals_dimensionless < self.ToverMmax)
            hlmT ={}
            taper_end_duration =None
            if rom_taper_end:
                taper_end_duration =40.0
            if 'Tidal' in self.param:
                time, hlmT_dimensionless_narrow,dym = self.sur(params_here[0],params_here[1],params_here[2],times=tvals_dimensionless[indx_ok],taper_end_duration=taper_end_duration,f_low=None,tidal_opts=params_here[5])
            else:
                time, hlmT_dimensionless_narrow,dym = self.sur(params_here[0],params_here[1],params_here[2],times=tvals_dimensionless[indx_ok],taper_end_duration=taper_end_duration,f_low=0)
#            hlmT_dimensionless_narrow = self.sur(params_here[0],params_here[1],params_here[2],dt=P.deltaT,f_low=0,mode_list=self.modes_available,M=params_here[3],dist_mpc=params_here[4],tidal_opts=params_here[5])
#            hlmT_dimensionless_narrow = self.sur(params_here,times=tvals_dimensionless[indx_ok],f_low=0,taper_end_duration=taper_end_duration)
            # Build taper for start
            taper_start_window = np.ones(len(hlmT_dimensionless_narrow[(2,2)]))
            if rom_taper_start or P.taper  != lalsimutils.lsu_TAPER_NONE:
                print " HybSur: Preparing manual tapering for first ~ 12 cycles "
                # Taper the start of the waveform by considering a fixed number of cyles for 2,2 mode
                # Taper window must be applied later on, as we want to measure the starting frequency
                # mode-by-mode
                fM_start_22_mode = gwtools.find_instant_freq(np.real(hlmT_dimensionless_narrow[(2,2)]),\
                    np.imag(hlmT_dimensionless_narrow[(2,2)]),tvals_dimensionless[indx_ok])
                taper_start_duration = (2.0 * np.pi) / np.abs(fM_start_22_mode) # about (2pi)^2 radians from 22 mode
                taper_start_window = gwtools.gwutils.window(tvals_dimensionless[indx_ok],\
                    tvals_dimensionless[indx_ok][0], tvals_dimensionless[indx_ok][0]+taper_start_duration,\
                    rolloff=0, windowType='planck')
                taper_fraction = taper_start_duration/(tvals_dimensionless[indx_ok][-1] - tvals_dimensionless[indx_ok][0])
                print "Taper duration "+str(taper_start_duration)+" (M). Fraction of signal: "+str(taper_fraction)

            for mode in self.modes_available:
                hlmT_dimensionless[mode] = np.zeros(len(tvals_dimensionless),dtype=complex)
                if mode[1]<0 and self.reflection_symmetric:   
                    # Perform reflection symmetry
                    mode_alt = (mode[0],-mode[1])
                    hlmT_dimensionless[mode][indx_ok] = np.power(-1, mode[0])*np.conj( taper_start_window* hlmT_dimensionless_narrow[mode_alt])
                else:
                    hlmT_dimensionless[mode][indx_ok] = taper_start_window* hlmT_dimensionless_narrow[mode]

        # Loop over all modes in the system
        for mode in self.modes_available: # loop over modes
          if mode[0] <= Lmax:
            # Copy into a new LIGO time series object
            wfmTS = lal.CreateCOMPLEX16TimeSeries("h", lal.LIGOTimeGPS(0.), 0., deltaT, lalsimutils.lsu_DimensionlessUnit, npts)
            wfmTS.data.data *=0;  # init - a sanity check


            # Option 1: Use the surrogate functions themselves
            if not use_basis:
                params_here = self.parameter_convert[mode](P)
                # Option 0: Use NRSur7dsq approach (i.e., use the stored hlmT data )
                if  not self.single_mode_sur: # 'NRSur7d' in self.param:
                    wfmTS.data.data = m_total_s/distance_s * hlmT_dimensionless[mode]
                else:
                    t_phys, hp_dim, hc_dim = self.post_dict[mode](*self.sur_dict[mode](q=params_here,samples=tvals/m_total_s + self.ToverM_peak))  # center time values AT PEAK
                    wfmTS.data.data =  m_total_s/distance_s * (hp_dim + 1j*hc_dim)  # ARGH! Is this consistent across surrogates?
                # Zero out data after tmax and before tmin
                wfmTS.data.data[tvals/m_total_s<self.ToverMmin-self.ToverM_peak] = 0
                wfmTS.data.data[tvals/m_total_s>self.ToverMmax-self.ToverM_peak] = 0
                # Apply taper *at times where ROM waveform starts*
                if P.taper:
                    n0 = np.argmin(np.abs(tvals)) # origin of time; this will map to the peak
                    # start taper
                    nstart = n0+ int((self.ToverMmin-self.ToverM_peak)*m_total_s/deltaT)        # note offset for tapering. Assumes time interval longer than requested...for NRHyb, ToverMmin does NOT relate to returned array size, beware!
                    nstart = np.max([0,nstart]) # in principle, waveform returned could be massively longer than available time
                    if nstart >0:
                        ntaper = int( (self.ToverM_peak-self.ToverMmin)*m_total_s/deltaT*0.1)  # SHOULD BE set by a few Hz in time fixed 1% of waveform length
                    else:
                        ntaper = int(0.05*len(tvals))   # force 1% length taper
                    vectaper= 0.5 - 0.5*np.cos(np.pi*np.arange(ntaper)/(1.*ntaper))
                    if rosDebug:
                        print " Tapering ROM hlm(t) for ", mode, " over range ", nstart, nstart+ntaper, " or time offset ", nstart*deltaT, " and window ", ntaper*deltaT
                        print  len(vectaper), ntaper, n0,nstart, len(tvals), wfmTS.data.length
                    wfmTS.data.data[nstart:nstart+ntaper]*=vectaper

                    # end taper
                    ntaper = int( (self.ToverMmax-self.ToverM_peak)*m_total_s/deltaT*0.1)  # SHOULD BE set by a few Hz in time fixed 1% of waveform length
                    nend = n0 + int((self.ToverMmax-self.ToverM_peak)*m_total_s/deltaT)
                    vectaper= 0.5 - 0.5*np.cos(np.pi* (1-np.arange(ntaper)/(1.*ntaper)))
                    wfmTS.data.data[-ntaper+nend:nend]*=vectaper
                    if rosDebug:
                        print " Tapering ROM hlm(t) end for ", mode, " over range ", nend-ntaper, nend, " or time offset ", nend*deltaT, ntaper*deltaT


            else:
                # Option 2: reconstruct with basis.  Note we will previously have called the basis function generator.
                # select index list
                indx_list_ok = [indx for indx in coefs.keys()  if indx[0]==mode[0] and indx[1]==mode[1]]
                if rosDebug:
                    print " To reconstruct ", mode, " using the following basis entries ",  indx_list_ok
                fudge_factor = 1    # FIXME, TERRIBLE!!!
                for indx in indx_list_ok: # loop over basis
#                    print " Adding " , bT[indx].data.length, wfmTS.data.length, indx, " of " ,len(coefs), (coefs[indx]*bT[indx].data.data)[len(bT[indx].data.data)/2]
                    wfmTS.data.data +=  fudge_factor*m_total_s/distance_s * np.array(coefs[indx]*bT[indx].data.data,dtype=np.cfloat)

                # CLEAN UP
                # Zero out data after tmax and before tmin
                wfmTS.data.data[tvals/m_total_s<self.ToverMmin-self.ToverM_peak] = 0
                wfmTS.data.data[tvals/m_total_s>self.ToverMmax-self.ToverM_peak] = 0
                # Apply taper *at times where ROM waveform starts*
                if P.taper:
                    n0 = np.argmin(np.abs(tvals)) # origin of time; this will map to the peak
                    # start taper
                    nstart = n0+int((self.ToverMmin-self.ToverM_peak)*m_total_s/deltaT)        # note offset for tapering
                    ntaper = int( (self.ToverM_peak-self.ToverMmin)*m_total_s/deltaT*0.1)  # SHOULD BE set by a few Hz in time fixed 1% of waveform length
                    vectaper= 0.5 - 0.5*np.cos(np.pi*np.arange(ntaper)/(1.*ntaper))
                    print " Tapering ROM hlm(t) for ", mode, " over range ", nstart, nstart+ntaper, " or time offset ", nstart*deltaT, " and window ", ntaper*deltaT
                    wfmTS.data.data[nstart:nstart+ntaper]*=vectaper

                    # end taper
                    ntaper = int( (self.ToverMmax-self.ToverM_peak)*m_total_s/deltaT*0.1)  # SHOULD BE set by a few Hz in time fixed 1% of waveform length
                    nend = n0 + int((self.ToverMmax-self.ToverM_peak)*m_total_s/deltaT)
                    vectaper= 0.5 - 0.5*np.cos(np.pi* (1-np.arange(ntaper)/(1.*ntaper)))
                    wfmTS.data.data[-ntaper+nend:nend]*=vectaper
                    print " Tapering ROM hlm(t) end for ", mode, " over range ", nend-ntaper, nend, " or time offset ", nend*deltaT, ntaper*deltaT



            # Set the epoch for the time series correctly: should have peak near center of series by construction
            # MUST USE TIME CENTERING INFO FROM SURROGATE
            wfmTS.epoch = -deltaT*wfmTS.data.length/2

            # Store the resulting mode
            hlmT[mode] = wfmTS

            # if the 22 mode, use to identify the natural frequency.  Can afford to be a bit sloppy, since only used for hybridization
            if hybrid_use and hybrid_time == None and mode[0]==2 and mode[1]==2 :
                if hybrid_frequency:
                    phase_vals = np.angle(wfmTS.data.data)
                    phase_vals = lalsimutils.unwind_phase(phase_vals)
                    datFreqReduced = (-1)* (np.roll(phase_vals,-1) - phase_vals)/deltaT/(2*np.pi)   # Hopefully this is smooth and monotonic
                    indx_max = np.argmax(np.abs(wfmTS.data.data))
                    t_max_location = tvals[indx_max]
                    indx_ok =  np.logical_and(np.abs(datFreqReduced) < hybrid_frequency*1.001, tvals < t_max_location)  # Use peak time to insure matching occurs early on
                    tvals_ok = tvals[indx_ok]
                    f_ok = datFreqReduced[indx_ok]
                    hybrid_time_viaf = tvals_ok[np.argmax(f_ok)]   # rely on high sampling rate. No interpolation!
#                    if verbose:
#                        print " NR catalog: revising hybridization time from 22 mode  to ", hybrid_time_viaf, " given frequency ", hybrid_frequency

        # hybridize
        if hybrid_use:
            my_hybrid_time = hybrid_time_viaf
#            HackRoundTransverseSpin(self.P) # Hack, sub-optimal
            if my_hybrid_time == None:
                my_hybrid_time = -0.5*self.estimateDurationSec(P)  # note fmin is not used. Note this is VERY conservative
            if verbose:
                print "  hybridization performed for ", self.group, self.param, " at time ", my_hybrid_time
            P.deltaT = deltaT # sanity
            # HACK: round digits, so I can get a spin-aligned approximant if I need it
            hlmT_hybrid = LALHybrid.hybridize_modes(hlmT,P,hybrid_time_start=my_hybrid_time,hybrid_method=hybrid_method)
            return hlmT_hybrid
        else:
            if rosDebug:
                print " ------ NO HYBRIDIZATION PERFORMED AT LOW LEVEL (=automatic) for ", self.group, self.param, "----- "
        return hlmT

    def hlmoff(self, P,force_T=False, deltaT=1./16384, time_over_M_zero=0.,use_basis=False,Lmax=np.inf,**kwargs):
        """
        hlmoff takes fourier transforms of LAL timeseries generated from hlmoft.
        All modes have physical units, appropriate to a physical signal.
        """
        hlmF ={}
        hlmT = self.hlmoft(P,force_T=force_T,deltaT=deltaT,time_over_M_zero=time_over_M_zero,use_basis=use_basis,Lmax=Lmax,**kwargs)
        for mode in hlmT.keys():
            wfmTS=hlmT[mode]
            # Take the fourier transform
            wfmFD = lalsimutils.DataFourier(wfmTS)  # this creates a new f series for *each* call.
            # Store the resulting mode
            hlmF[mode] = wfmFD
        return hlmF

    def conj_hlmoff(self, P,force_T=False, deltaT=1./16384, time_over_M_zero=0.,use_basis=False,Lmax=np.inf,**kwargs):
        """
        conj_hlmoff takes fourier transforms of LAL timeseries generated from hlmoft, but after complex conjugation.
        All modes have physical units, appropriate to a physical signal.
        """
        hlmF ={}
        hlmT = self.hlmoft(P,force_T=force_T,deltaT=deltaT,time_over_M_zero=time_over_M_zero,use_basis=use_basis,Lmax=Lmax,**kwargs)
        for mode in hlmT.keys():
            wfmTS=hlmT[mode]
            wfmTS.data.data = np.conj(wfmTS.data.data)  # complex conjugate
            # Take the fourier transform
            wfmFD = lalsimutils.DataFourier(wfmTS)  # this creates a new f series for *each* call.
            # Store the resulting mode
            hlmF[mode] = wfmFD
        return hlmF


class WaveformMode:
    """
    Class representing a dimensionless timeseries X(\tau=t/M)
    """

###
### FFT syntatic sugar
###
def DataFourierNumpy(wfComplex):    # assume (n,2) size array of [tvals, g(t)]; return fvals, tilde(g). 
    # FFT
    # Unroll -- it will save me time later if we are continuous
    T = wfComplex[-1,0] - wfComplex[0,0]
    n = len(wfComplex)
    dt = T/n
    gtilde = np.fft.fft(wfComplex[:,1])*dt   
    # Option 1: Raw pairing, not caring about continuity in the frequency array.
#    fvals = 1./T* np.array([ k if  k<=n/2 else k-n for k in np.arange(n)]) 
    # Option 2: Unroll
    gtilde = np.roll(gtilde, n/2)
    fvals = 1./T*(np.arange(n) - n/2+1)     # consistency with DataFourier: \int dt e(i2\pi f t), note REVERSING
    wfComplexF = np.array([fvals,gtilde[::-1]]).T
    return wfComplexF

def DataInverseFourierNumpy(wfComplex):    # assume (n,2) size array of [fvals, gtilde(f)]; return tvals, g
#    print "NAN check ", wfComplex[np.isnan(wfComplex[:,1]),1]
    df = wfComplex[1,0] - wfComplex[0,0]
    n = len(wfComplex)
    T = 1./df
    dt = T/n
    # Undo the roll, then perform the FFT
    datReversed = wfComplex[:,1][::-1]
    g = np.fft.ifft(np.roll(datReversed, -n/2))*n*df # undo the reverse and roll
    # Assign correct time values.  
    #    - Note the zero of time is now centered in the array -- we don't carry a time reference with us.
    tvals = dt*(np.arange(n) -n/2+1)
    wfComplexT = np.array([tvals,g]).T
    return wfComplexT

###
### Mode syntactic sugar
###
def RawGetModePeakTime(wfMode):   # assumed applied to complex data sequence
    nmax = np.argmax(np.abs(wfMode[:,1]))
    return np.real(wfMode[nmax][0])



###
### Plotting syntactic sugar
###

# http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.plot    
mode_line_style= {}
mode_line_style[(2,2)] = 'r'
mode_line_style[(2,1)] = 'b'
mode_line_style[(2,-1)] = 'b'
mode_line_style[(2,-2)] = 'r'
mode_line_style[(3,3)] = 'r-'
mode_line_style[(3,-3)] = 'r-'
mode_line_style[(3,1)] = 'b-'
mode_line_style[(3,-1)] = 'b-'

mode_line_style[(2,0)] = 'g'


for l in np.arange(2,5):
    for m in np.arange(-l,l+1):
        if not ((l,m) in mode_line_style.keys()):
            mode_line_style[(l,m)] = 'k'




#!/opt/local/bin/python

# This code was used to study the properties of the integral over distance
# in our parameter estimation code. The results of the study are documented
# in the notes overview.tex in the last section of the appendix. I have not
# assessed the speedup relative to a monte carlo integrator. I did use a
# a smart gridded itegrator which suggests that a lot of samples are needed
# get good accuracy. A comparison of the cost to get accurate results from 
# the integrator versus using the approximation gives huge speed ups ~40. But
# this must be taken with a grain of salt since a smart MC integrator may not
# need so many samples in distance to get the same accuracy. 

import numpy as np
import pylab as pl
from scipy.integrate import simps
from scipy.special import erf

pl.rc('text', usetex=True)

npoints=1024*8*8
sigmahat = np.array([0.0001,0.001,0.01,0.10,0.2,0.3,0.4,0.5])
rho = np.linspace(-5.0, 15.0, 1000)
rho = rho[:,np.newaxis]

for a in sigmahat:
  rhosigma = rho * a
  s2 = a * a
 
  ymax = 2*2.0 * np.max(rho) / a
  ymid = 0.25 * ymax # 0.5 / a
  
  #
  # Marginalize over distance numerically
  #
  y = np.linspace(1.0,ymid,npoints)
  dy = y[1]-y[0]
  y4 = y * y
  y4 = y4 * y4
  evidence = ( np.sum( np.exp( ( rhosigma - 0.5 * s2 * y ) * y ) / y4 , axis=1 ) * dy ) 
  y = np.linspace(ymid,ymax,npoints)
  dy = y[1]-y[0]
  y4 = y * y
  y4 = y4 * y4
  evidence = evidence + ( np.sum( np.exp( ( rhosigma - 0.5 * s2 * y ) * y ) / y4 , axis=1 ) * dy ) 

  #
  # Construct an approximation to the evidence
  #
  xplus = ( rho - 1.0 )
  xminus = ( rho - 10.0 )
  fplus = 0.5*(1.0 + np.tanh( xplus ))
  fminus = 0.5*(1.0 - np.tanh( xminus ) )

  # High snr approximation
  sgn = np.sign( rho - a )
  factor =  0.5 * (np.sign( rho - 2.0 ) + 1) * ( a * a * a ) * ( np.sqrt( np.pi / 2.0 ) * (1.0 + sgn * erf( sgn * (rho - a)/np.sqrt(2.0) ) ) / np.power(rho, 4.0) - 4.0 * np.exp( - 0.5 * a * a + a * rho - 0.5 * rho * rho ) / np.power(rho, 5.0) )
  evidence1 = factor * np.exp( ( ( 0.5 *rho * rho ) ) )
 
  # Low snr approximation
  evidence1 = evidence1 + 0.3333333333 * np.exp( ( - 0.5 * a * a + rho * a ) )

  #
  # Plot the evidence and the approximation
  #
  pl.figure(1)
  mylabel = "$\hat{\sigma} = %0.4f$" % a
  pl.semilogy(rho, evidence, '-', label=mylabel)
  pl.semilogy(rho, evidence1, 'o')

  # 
  # Plot the fractional error as a function of evidence
  #
  pl.figure(2)
  ratio = evidence/evidence1.ravel()
  pl.semilogx(evidence, np.abs(ratio), label=mylabel)

  print "This is a visual pacifier" 

pl.figure(1)
pl.legend(loc='upper left')
pl.xlabel(r'$\rho$', fontsize=16)
pl.ylabel(r'$I/r_{\mathrm{max}^3}$', fontsize=16)
pl.grid(True)
pl.savefig("i-v-rho.png")
pl.xlim(-1.0,10.0)
pl.ylim(1.0e-1,1.0e10)
pl.savefig("i-v-rho-zoom.png")

pl.figure(2)
pl.legend(loc='upper right')
pl.xlabel(r'$I/r_{\mathrm{max}}^3$', fontsize=16)
pl.ylabel(r'$I/I_{\mathrm{approx}}$', fontsize=16)
pl.grid(True)
pl.savefig("i-error-v-i.png")
pl.xlim(1.0e-1,1.0e10)
pl.savefig("i-error-v-i-zoom.png")




#! /usr/bin/env python

# plot_posterior_corner.py
#
#  --posterior-file f1 --posterior-label n1 --posterior-file f2 --posterior-label n2 ...
#  --parameter p1 --parameter p2 ...
#
# EXAMPLE
#    python plot_posterior_corner.py --posterior-file downloads/TidalP4.dat --parameter lambda1 --parameter lambda2 --parameter mc
#   python plot_posterior_corner.py --parameter mc --parameter eta --posterior-file G298048/production_C00_cleaned_TaylorT4/posterior-samples.dat  --parameter lambdat
#
# USAGE
#    - hardcoded list of colors, used in order, for multiple plots
#

import RIFT.lalsimutils as lalsimutils
import lal
import numpy as np
import argparse

eos_param_names = ['logp1', 'gamma1','gamma2', 'gamma3', 'R1_km', 'R2_km']


try:
    import matplotlib
    print " Matplotlib backend ", matplotlib.get_backend()
    if matplotlib.get_backend() is 'agg':
        fig_extension = '.png'
        bNoInteractivePlots=True
    else:
	matplotlib.use('agg')
        fig_extension = '.png'
        bNoInteractivePlots =True
    from matplotlib import pyplot as plt
    bNoPlots=False
except:
    print " Error setting backend"

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.lines as mlines
import corner

import RIFT.misc.our_corner as our_corner
try:
    import RIFT.misc.bounded_kde as bounded_kde
except:
    print " -No 1d kdes- "

dpi_base=200
legend_font_base=16
rc_params = {'backend': 'ps',
             'axes.labelsize': 11,
             'axes.titlesize': 10,
             'font.size': 11,
             'legend.fontsize': legend_font_base,
             'xtick.labelsize': 11,
             'ytick.labelsize': 11,
             #'text.usetex': True,
             'font.family': 'Times New Roman'}#,
             #'font.sans-serif': ['Bitstream Vera Sans']}#,
plt.rcParams.update(rc_params)


print " WARNINGS : BoundedKDE class can oversmooth.  Need to edit options for using this class! "

def render_coord(x,logscale=False):
    if x in lalsimutils.tex_dictionary.keys():
        mystr= lalsimutils.tex_dictionary[x]
        if logscale:
            mystr=mystr.lstrip('$')
            mystr = "$\log_{10}"+mystr
            return mystr
        else:
            return mystr
    if 'product(' in x:
        a=x.replace(' ', '') # drop spaces
        a = a[:len(a)-1] # drop last
        a = a[8:]
        terms = a.split(',')
        exprs =map(render_coord, terms)
        exprs = map( lambda x: x.replace('$', ''), exprs)
        my_label = ' '.join(exprs)
        return '$'+my_label+'$'
    else:
        return x

def render_coordinates(coord_names,logparams=[]):
    return map(lambda x: render_coord(x,logscale=(x in logparams)), coord_names)


def add_field(a, descr):
    """Return a new array that is like "a", but has additional fields.

    Arguments:
      a     -- a structured numpy array
      descr -- a numpy type description of the new fields

    The contents of "a" are copied over to the appropriate fields in
    the new array, whereas the new fields are uninitialized.  The
    arguments are not modified.

    >>> sa = numpy.array([(1, 'Foo'), (2, 'Bar')], \
                         dtype=[('id', int), ('name', 'S3')])
    >>> sa.dtype.descr == numpy.dtype([('id', int), ('name', 'S3')])
    True
    >>> sb = add_field(sa, [('score', float)])
    >>> sb.dtype.descr == numpy.dtype([('id', int), ('name', 'S3'), \
                                       ('score', float)])
    True
    >>> numpy.all(sa['id'] == sb['id'])
    True
    >>> numpy.all(sa['name'] == sb['name'])
    True
    """
    if a.dtype.fields is None:
        raise ValueError, "`A' must be a structured numpy array"
    b = np.empty(a.shape, dtype=a.dtype.descr + descr)
    for name in a.dtype.names:
        b[name] = a[name]
    return b

remap_ILE_2_LI = {
 "s1z":"a1z", "s2z":"a2z", 
 "s1x":"a1x", "s1y":"a1y",
 "s2x":"a2x", "s2y":"a2y",
 "chi1_perp":"chi1_perp",
 "chi2_perp":"chi2_perp",
 "chi1":'a1',
 "chi2":'a2',
 "cos_phiJL": 'cos_phiJL',
 "sin_phiJL": 'sin_phiJL',
 "cos_theta1":'costilt1',
 "cos_theta2":'costilt2',
 "theta1":"tilt1",
 "theta2":"tilt2",
  "xi":"chi_eff", 
  "chiMinus":"chi_minus", 
  "delta":"delta", 
  "delta_mc":"delta", 
 "mtot":'mtotal', "mc":"mc", "eta":"eta","m1":"m1","m2":"m2",
  "cos_beta":"cosbeta",
  "beta":"beta",
  "LambdaTilde":"lambdat",
  "DeltaLambdaTilde": "dlambdat",
  "thetaJN":"theta_jn"}
remap_LI_to_ILE = { "a1z":"s1z", "a2z":"s2z", "chi_eff":"xi", "lambdat":"LambdaTilde", 'mtotal':'mtot', "distance":"dist", 'ra':'phi', 'dec':'theta',"phiorb":"phiref"}


def extract_combination_from_LI(samples_LI, p):
    """
    extract_combination_from_LI
      - reads in known columns from posterior samples
      - for selected known combinations not always available, it will compute them from standard quantities
    Unike version in ConstructIntrinsicPosterior, this code does not rely on ChooseWaveformParams to perform coordinate changes...
    """
    if p in samples_LI.dtype.names:  # e.g., we have precomputed it
        return samples_LI[p]
    if p in remap_ILE_2_LI.keys():
       if remap_ILE_2_LI[p] in samples_LI.dtype.names:
         return samples_LI[ remap_ILE_2_LI[p] ]
    # Return cartesian components of spin1, spin2.  NOTE: I may already populate these quantities in 'Add important quantities'
    if (p == 'chi_eff' or p=='xi') and 'a1z' in samples_LI.dtype.names:
        m1 = samples_LI['m1']
        m2 = samples_LI['m2']
        a1z = samples_LI['a1z']
        a2z = samples_LI['a2z']
        return (m1 * a1z + m2*a2z)/(m1+m2)
    if p == 'chiz_plus':
        print " Transforming "
        if 'a1z' in samples_LI.dtype.names:
            return (samples_LI['a1z']+ samples_LI['a2z'])/2.
        if 'theta1' in samples_LI.dtype.names:
            return (samples_LI['a1']*np.cos(samples_LI['theta1']) + samples_LI['a2']*np.cos(samples_LI['theta2']) )/2.
#        return (samples_LI['a1']+ samples_LI['a2'])/2.
    if p == 'chiz_minus':
        print " Transforming "
        if 'a1z' in samples_LI.dtype.names:
            return (samples_LI['a1z']- samples_LI['a2z'])/2.
        if 'theta1' in samples_LI.dtype.names:
            return (samples_LI['a1']*np.cos(samples_LI['theta1']) - samples_LI['a2']*np.cos(samples_LI['theta2']) )/2.
#        return (samples_LI['a1']- samples_LI['a2'])/2.
    if  'theta1' in samples_LI.dtype.names:
        if p == 's1x':
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) * np.cos( samples_LI['phi1'])
        if p == 's1y' :
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) * np.sin( samples_LI['phi1'])
        if p == 's2x':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) * np.cos( samples_LI['phi2'])
        if p == 's2y':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) * np.sin( samples_LI['phi2'])
        if p == 'chi1_perp' :
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) 
        if p == 'chi2_perp':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) 
    elif  'tilt1' in samples_LI.dtype.names:
        if p == 'chi1_perp' :
            return samples_LI["a1"]*np.sin(samples_LI[ 'tilt1']) 
        if p == 'chi2_perp':
            return samples_LI["a2"]*np.sin(samples_LI[ 'tilt2']) 
    else:  # aligned
        if p == 'chi1_perp' :
            return np.zeros(len(samples_LI["m1"])) 
        if p == 'chi2_perp':
            return np.zeros(len(samples_LI["m1"])) 

    if 'lambdat' in samples_LI.dtype.names:  # LI does sampling in these tidal coordinates
        lambda1, lambda2 = lalsimutils.tidal_lambda_from_tilde(samples_LI["m1"], samples_LI["m2"], samples_LI["lambdat"], samples_LI["dlambdat"])
        if p == "lambda1":
            return lambda1
        if p == "lambda2":
            return lambda2
    if p == 'delta' or p=='delta_mc':
        return (samples_LI['m1']  - samples_LI['m2'])/((samples_LI['m1']  + samples_LI['m2']))
    # Return cartesian components of Lhat
    if p == 'product(sin_beta,sin_phiJL)':
        return np.sin(samples_LI[ remap_ILE_2_LI['beta'] ]) * np.sin(  samples_LI['phi_jl'])
    if p == 'product(sin_beta,cos_phiJL)':
        return np.sin(samples_LI[ remap_ILE_2_LI['beta'] ]) * np.cos(  samples_LI['phi_jl'])

    if p == 'mc':
        m1v= samples["m1"]
        m2v = samples["m2"]
        return lalsimutils.mchirp(m1v,m2v)
    if p == 'eta':
        m1v= samples["m1"]
        m2v = samples["m2"]
        return lalsimutils.symRatio(m1v,m2v)

    # Backup : access lambdat if not present
    if (p == 'lambdat' or p=='dlambdat') and 'lambda1' in samples.dtype.names:
        Lt,dLt = lalsimutils.tidal_lambda_tilde(samples['m1'], samples['m2'],  samples['lambda1'], samples['lambda2'])
        if p=='lambdat':
            return Lt
        if p=='dlambdat':
            return dLt

    if p == "q"  and 'm1' in samples.dtype.names:
        return samples["m2"]/samples["m1"]

    print " No access for parameter ", p, " in ", samples.dtype.names
    return np.zeros(len(samples_LI['m1']))  # to avoid causing a hard failure



# Parse arguments
parser = argparse.ArgumentParser()
parser.add_argument("--posterior-file",action='append',help="filename of *.dat file [standard LI output]")
parser.add_argument("--truth-file",type=str, help="file containing the true parameters")
parser.add_argument("--posterior-distance-factor",action='append',help="Sequence of factors used to correct the distances")
parser.add_argument("--truth-event",type=int, default=0,help="file containing the true parameters")
parser.add_argument("--composite-file",action='append',help="filename of *.dat file [standard ILE intermediate]")
parser.add_argument("--flag-tides-in-composite",action='store_true',help='Required, if you want to parse files with tidal parameters')
parser.add_argument("--posterior-label",action='append',help="label for posterior file")
parser.add_argument("--posterior-color",action='append',help="color and linestyle for posterior. PREPENDED onto default list, so defaults exist")
parser.add_argument("--posterior-linestyle",action='append',help="color and linestyle for posterior. PREPENDED onto default list, so defaults exist")
parser.add_argument("--parameter", action='append',help="parameter name (ILE). Note source-frame masses are only natively supported for LI")
parser.add_argument("--parameter-log-scale",action='append',help="Put this parameter in log scale")
parser.add_argument("--change-parameter-label", action='append',help="format name=string. Will be wrapped in $...$")
parser.add_argument("--use-legend",action='store_true')
parser.add_argument("--use-title",default=None,type=str)
parser.add_argument("--use-smooth-1d",action='store_true')
parser.add_argument("--plot-1d-extra",action='store_true')
parser.add_argument("--pdf",action='store_true',help="Export PDF plots")
parser.add_argument("--mc-range",default=None,help='List for mc range. Default is None')
parser.add_argument("--ci-list",default=None,help='List for credible intervals. Default is 0.95,0.9,0.68')
parser.add_argument("--quantiles",default=None,help='List for 1d quantiles intervals. Default is 0.95,0.05')
parser.add_argument("--chi-max",default=1,type=float)
parser.add_argument("--lambda-plot-max",default=2000,type=float)
parser.add_argument("--lnL-cut",default=None,type=float)
parser.add_argument("--sigma-cut",default=0.4,type=float)
opts=  parser.parse_args()
if opts.posterior_file is None:
    print " No input files "
    import sys
    sys.exit(0)
if opts.pdf:
    fig_extension='.pdf'

truth_P_list = None
P_ref = None
if opts.truth_file:
    print " Loading true parameters from  ", opts.truth_file
    truth_P_list  =lalsimutils.xml_to_ChooseWaveformParams_array(opts.truth_file)
    P_ref = truth_P_list[opts.truth_event]
#    P_ref.print_params()

if opts.change_parameter_label:
  for name, new_str in map( lambda c: c.split("="),opts.change_parameter_label):
      if name in lalsimutils.tex_dictionary:
          lalsimutils.tex_dictionary[name] = "$"+new_str+"$"

special_param_ranges = {
  'q':[0,1],
  'eta':[0,0.25],
  'a1z':[-opts.chi_max,opts.chi_max],
  'a2z':[-opts.chi_max,opts.chi_max],
  'chi_eff': [-opts.chi_max,opts.chi_max],  # this can backfire for very narrow constraints
  'lambda1':[0,4000],
  'lambda2':[0,4000],
  'lambdat':[0,4000]
}
if opts.mc_range:
    special_param_ranges['mc'] = eval(opts.mc_range)
    print " mc range ", special_param_ranges['mc']
    


# Parameters
param_list = opts.parameter

# Legend
color_list=['black', 'red', 'green', 'blue','yellow','C0','C1','C2','C3']
if opts.posterior_color:
    color_list  =opts.posterior_color + color_list
else:
    color_list += len(opts.posterior_file)*['black']
linestyle_list = ['-' for k in color_list]
if opts.posterior_linestyle:
    linestyle_list = opts.posterior_linestyle + linestyle_list
#linestyle_remap_contour  = {":", 'dotted', '-'
posterior_distance_factors = np.ones(len(opts.posterior_file))
if opts.posterior_distance_factor:
    for indx in np.arange(len(opts.posterior_file)):
        posterior_distance_factors[indx] = float(opts.posterior_distance_factor[indx])

line_handles = []
corner_legend_location=None; corner_legend_prop=None
if opts.use_legend and opts.posterior_label:
    n_elem = len(opts.posterior_file)
    for indx in np.arange(n_elem):
        my_line = mlines.Line2D([],[],color=color_list[indx],linestyle=linestyle_list[indx],label=opts.posterior_label[indx])
        line_handles.append(my_line)

    corner_legend_location=(0.7, 1.0)
    corner_legend_prop = {'size':8}
# https://stackoverflow.com/questions/7125009/how-to-change-legend-size-with-matplotlib-pyplot
#params = {'legend.fontsize': 20, 'legend.linewidth': 2}
#plt.rcParams.update(params)


# Import
posterior_list = []
posteriorP_list = []
label_list = []
# Load posterior files
if opts.posterior_file:
 for fname in opts.posterior_file:
    samples = np.genfromtxt(fname,names=True)
    if not 'mtotal' in samples.dtype.names and 'mc' in samples.dtype.names:  # raw LI samples use 
        q_here = samples['q']
        eta_here = q_here/(1+q_here)
        mc_here = samples['mc']
        mtot_here = mc_here / np.power(eta_here, 3./5.)
        m1_here = mtot_here/(1+q_here)
        samples = add_field(samples, [('mtotal', float)]); samples['mtotal'] = mtot_here
        samples = add_field(samples, [('eta', float)]); samples['eta'] = eta_here
        samples = add_field(samples, [('m1', float)]); samples['m1'] = m1_here
        samples = add_field(samples, [('m2', float)]); samples['m2'] = mtot_here * q_here/(1+q_here)

    if (not 'theta1' in samples.dtype.names)  and ('a1x' in samples.dtype.names):  # probably does not have polar coordinates
        chiperp_here = np.sqrt( samples['a1x']**2+ samples['a1y']**2)
        chi1_here = np.sqrt( samples['a1z']**2 + chiperp_here**2)
        theta1_here = np.arctan( samples['a1z']/chiperp_here)
        phi1_here = np.angle(samples['a1x']+1j*samples['a1y'])
        samples = add_field(samples, [('chi1', float)]); samples['chi1'] = chi1_here
        samples = add_field(samples, [('theta1', float)]); samples['theta1'] = theta1_here
        samples = add_field(samples, [('phi1', float)]); samples['phi1'] = phi1_here
        
        # we almost certainly use standard
        chi1_perp = np.sqrt(samples['a1x']**2 + samples['a1y']**2)
        chi2_perp = np.sqrt(samples['a2x']**2 + samples['a2y']**2)
        samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
        samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp
        
    elif "theta1" in samples.dtype.names:
        a1x_dat = samples["a1"]*np.sin(samples["theta1"])*np.cos(samples["phi1"])
        a1y_dat = samples["a1"]*np.sin(samples["theta1"])*np.sin(samples["phi1"])
        chi1_perp = samples["a1"]*np.sin(samples["theta1"])

        a2x_dat = samples["a2"]*np.sin(samples["theta2"])*np.cos(samples["phi2"])
        a2y_dat = samples["a2"]*np.sin(samples["theta2"])*np.sin(samples["phi2"])
        chi2_perp = samples["a2"]*np.sin(samples["theta2"])

                                      
        samples = add_field(samples, [('a1x', float)]);  samples['a1x'] = a1x_dat
        samples = add_field(samples, [('a1y', float)]); samples['a1y'] = a1y_dat
        samples = add_field(samples, [('a2x', float)]);  samples['a2x'] = a2x_dat
        samples = add_field(samples, [('a2y', float)]);  samples['a2y'] = a2y_dat
        samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
        samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp
        if not 'chi_eff' in samples.dtype.names:
            samples = add_field(samples, [('chi_eff',float)]); samples['chi_eff'] = (samples["m1"]*samples["a1z"]+samples["m2"]*samples["a2z"])/(samples["m1"]+samples["m2"])
 
    elif 'a1x' in samples.dtype.names:
        chi1_perp = np.sqrt(samples['a1x']**2 + samples['a1y']**2)
        chi2_perp = np.sqrt(samples['a2x']**2 + samples['a2y']**2)
        samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
        samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp

    if 'lambda1' in samples.dtype.names and not ('lambdat' in samples.dtype.names):
        Lt,dLt = lalsimutils.tidal_lambda_tilde(samples['m1'], samples['m2'],  samples['lambda1'], samples['lambda2'])
        samples = add_field(samples, [('lambdat', float)]); samples['lambdat'] = Lt
        samples = add_field(samples, [('dlambdat', float)]); samples['dlambdat'] = dLt


    if 'chi1_perp' in samples.dtype.names:
        # impose Kerr limit, if neede
        npts = len(samples["m1"])
        indx_ok =np.arange(npts)
        chi1_squared = samples['chi1_perp']**2 + samples["a1z"]**2
        chi2_squared = samples["chi2_perp"]**2 + samples["a2z"]**2
        indx_ok = np.logical_and(chi1_squared < opts.chi_max ,chi2_squared < opts.chi_max)
        npts_out = np.sum(indx_ok)
        new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
        for name in samples.dtype.names:
            new_samples[name] = samples[name][indx_ok]
        samples = new_samples


    # Save samples
    posterior_list.append(samples)

    # Continue ... rest not used at present
    continue

    # Populate a P_list with the samples, so I can perform efficient conversion for plots
    # note only the DETECTOR frame properties are stored here
    P_list = []
    P = lalsimutils.ChooseWaveformParams()
    for indx in np.arange(len(samples["m1"])):
        P.m1 = samples["m1"][indx]*lal.MSUN_SI
        P.m2 = samples["m2"][indx]*lal.MSUN_SI
        P.s1x = samples["a1x"][indx]
        P.s1y = samples["a1y"][indx]
        P.s1z = samples["a1z"][indx]
        P.s2x = samples["a2x"][indx]
        P.s2y = samples["a2y"][indx]
        P.s2z = samples["a2z"][indx]
        if "lnL" in samples.keys():
            P.lnL = samples["lnL"][indx]   # creates a new field !
        else:
            P.lnL = -1
        # Populate other parameters as needed ...
        P_list.append(P)
    posteriorP_list.append(P_list)

for indx in np.arange(len(posterior_list)):
    samples = posterior_list[indx]
    fac = posterior_distance_factors[indx]
    if 'dist' in samples.dtype.names:
        samples["dist"]*= fac
    if 'distance' in samples.dtype.names:
        samples["distance"]*= fac

# Import
composite_list = []
field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","lnL", "sigmaOverL", "ntot", "neff")
if opts.flag_tides_in_composite:
    print " Reading composite file, assuming tide-based format "
    field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","lambda1", "lambda2", "lnL", "sigmaOverL", "ntot", "neff")
field_formats = [np.float32 for x in field_names]
composite_dtype = [ (x,float) for x in field_names] #np.dtype(names=field_names ,formats=field_formats)
# Load posterior files
if opts.composite_file:
 print opts.composite_file
 for fname in opts.composite_file[:1]:  # Only load the first one!
    print " Loading ... ", fname
    samples = np.loadtxt(fname,dtype=composite_dtype)  # Names are not always available
    if opts.sigma_cut >0:
        npts = len(samples["m1"])
        # strip NAN
        sigma_vals = samples["sigmaOverL"]
        good_sigma = sigma_vals < opts.sigma_cut
        npts_out = np.sum(good_sigma)
        if npts_out < npts:
            new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
            for name in samples.dtype.names:
                new_samples[name] = samples[name][good_sigma]
            samples = new_samples
    if opts.lnL_cut:
        npts = len(samples["m1"])
        # strip NAN
        lnL_vals = samples["lnL"]
        not_nan = np.logical_not(np.isnan(lnL_vals))
        npts_out = np.sum(not_nan)
        if npts_out < npts:
            new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
            for name in samples.dtype.names:
                new_samples[name] = samples[name][not_nan]
            samples = new_samples
        
        # apply cutoff
        indx_ok =np.arange(npts)
        lnL_max = np.max(samples["lnL"])
        print " lnL_max = ", lnL_max
        indx_ok = samples["lnL"]>lnL_max  -opts.lnL_cut
        npts_out = np.sum(indx_ok)
        new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
        for name in samples.dtype.names:
            new_samples[name] = samples[name][indx_ok]
        samples = new_samples
#    samples = np.recarray(samples.T,names=field_names,dtype=field_formats) #,formats=field_formats)
    # If no record names
    # Add mtotal, q, 
    samples=add_field(samples,[('mtotal',float)]); samples["mtotal"]= samples["m1"]+samples["m2"]; 
    samples=add_field(samples,[('q',float)]); samples["q"]= samples["m2"]/samples["m1"]; 
    samples=add_field(samples,[('mc',float)]); samples["mc"] = lalsimutils.mchirp(samples["m1"], samples["m2"])
    samples=add_field(samples,[('eta',float)]); samples["eta"] = lalsimutils.symRatio(samples["m1"], samples["m2"])
    samples=add_field(samples,[('chi_eff',float)]); samples["chi_eff"]= (samples["m1"]*samples["a1z"]+samples["m2"]*samples["a2z"])/(samples["mtotal"]); 
    chi1_perp = np.sqrt(samples['a1x']*samples["a1x"] + samples['a1y']**2)
    chi2_perp = np.sqrt(samples['a2x']**2 + samples['a2y']**2)
    samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
    samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp

    if ('lambda1' in samples.dtype.names):
        Lt,dLt = lalsimutils.tidal_lambda_tilde(samples['m1'], samples['m2'],  samples['lambda1'], samples['lambda2'])
        samples= add_field(samples, [('LambdaTilde',float), ('DeltaLambdaTilde',float),('lambdat',float),('dlambdat',float)])
        samples['LambdaTilde'] = samples['lambdat']= Lt
        samples['DeltaLambdaTilde'] = samples['dlambdat']= dLt


    print " Loaded samples from ", fname , len(samples["m1"])
    if True:
        # impose Kerr limit
        npts = len(samples["m1"])
        indx_ok =np.arange(npts)
        chi1_squared = samples["chi1_perp"]**2 + samples["a1z"]**2
        chi2_squared = samples["chi2_perp"]**2 + samples["a2z"]**2
        indx_ok = np.logical_and(chi1_squared < opts.chi_max ,chi2_squared < opts.chi_max)
        npts_out = np.sum(indx_ok)
        if npts_out < npts:
            print " Ok systems ", npts_out
            new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
            for name in samples.dtype.names:
                new_samples[name] = samples[name][indx_ok]
            samples = new_samples
            print " Stripped samples  from ", fname , len(samples["m1"])


    composite_list.append(samples)

    continue


## Plot posterior files

CIs = [0.95,0.9, 0.68]
if opts.ci_list:
    CIs = eval(opts.ci_list)  # force creation
quantiles_1d = [0.05,0.95]
if opts.quantiles:
    quantiles_1d=eval(opts.quantiles)

# Generate labels
if opts.parameter_log_scale is None:
    opts.parameter_log_scale = []
labels_tex = render_coordinates(opts.parameter,logparams=opts.parameter_log_scale)#map(lambda x: tex_dictionary[x], coord_names)

fig_base= None
# Create figure workspace for 1d plots
fig_1d_list = []
fig_1d_list_cum = []
#fig_1d_list_ids = []
if opts.plot_1d_extra:
    for indx in np.arange(len(opts.parameter))+5:
        fig_1d_list.append(plt.figure(indx))
        fig_1d_list_cum.append(plt.figure(indx+len(opts.parameter)))
#        fig_1d_list_ids.append(indx)
    plt.figure(1)


# Find parameter ranges
x_range = {}
range_list = []
if opts.posterior_file:
 for param in opts.parameter:
    xmax_list = []
    xmin_list = []
    for indx in np.arange(len(posterior_list)):
        dat_here = None
        samples = posterior_list[indx]
        if param in samples.dtype.names:
            dat_here = samples[param]
        else:
            dat_here = extract_combination_from_LI(samples, param)
        if param in opts.parameter_log_scale:
            indx_ok = dat_here > 0
            dat_here= np.log10(dat_here[indx_ok])
        if len(dat_here) < 1:
            print " Failed to etract data ", param,  " from ", opts.posterior_file[indx]

        # extend the limits, so we have *extremal* limits 
        xmax_list.append(np.max(dat_here))
        xmin_list.append(np.min(dat_here))
    x_range[param] = np.array([np.min(xmin_list), np.max(xmax_list)])  # give a small buffer
#    if param == 'chi_eff':
#        x_range[param] -= 0.1*np.sign([-1,1])*(x_range[param]+np.array([-1,1]))
            
    if param in special_param_ranges:
        x_range[param] = special_param_ranges[param]

    if param in ['lambda1', 'lambda2', 'lambdat']:
        x_range[param][1] = opts.lambda_plot_max

    range_list.append(x_range[param])
    print param, x_range[param]


my_cmap_values=None
for pIndex in np.arange(len(posterior_list)):
    samples = posterior_list[pIndex]
    sample_names = samples.dtype.names; sample_ref_name  = sample_names[0]
    # Create data for corner plot
    dat_mass = np.zeros( (len(samples[sample_ref_name]), len(labels_tex)) )
    my_cmap_values = color_list[pIndex]
    plot_range_list = []
    smooth_list =[]
    truths_here= None
    if opts.truth_file:
        truths_here = np.zeros(len(opts.parameter))
    for indx in np.arange(len(opts.parameter)):
        param = opts.parameter[indx]
        if param in samples.dtype.names:
            dat_mass[:,indx] = samples[param]
        else:
            dat_mass[:,indx] = extract_combination_from_LI(samples, param)

        if param in opts.parameter_log_scale:
            dat_mass[:,indx] = np.log10(dat_mass[:,indx])

        # Parameter ranges (duplicate)
        dat_here = np.array(dat_mass[:,indx])  # force copy ! I need to sort
        weights = np.ones(len(dat_here))*1.0/len(dat_here)
        if 'weights' in samples.dtype.names:
            weights = samples['weights']
        indx_sort= dat_here.argsort()
        dat_here = dat_here[indx_sort]
        weights =weights[indx_sort]
#
        dat_here.sort() # sort it
        xmin, xmax = x_range[param]
#            xmin = np.min([np.min( posterior_list[x][param]) for x in np.arange(len(posterior_list)) ]) # loop over all
        xmin  = np.min([xmin, np.mean(dat_here)  -4*np.std(dat_here)])
        xmax = np.max([xmax, np.mean(dat_here)  +4*np.std(dat_here)])
#            xmax  = np.max(dat_here)
        if param in special_param_ranges:
                xmin,xmax = special_param_ranges[param]
        plot_range_list.append((xmin,xmax))

        # smoothing list
        smooth_list.append(np.std(dat_here)/np.power(len(dat_here), 1./3))
        
        # truths
        if opts.truth_file:
            param_to_extract = param
            if param in remap_LI_to_ILE.keys():
                param_to_extract  = remap_LI_to_ILE[param]
            if param in eos_param_names:
                continue
            truths_here[indx] = P_ref.extract_param(param_to_extract)
            if param in [ 'mc', 'm1', 'm2', 'mtotal']:
                truths_here[indx] = truths_here[indx]/lal.MSUN_SI
            if param in ['dist', 'distance']:
                truths_here[indx] = truths_here[indx]/lal.PC_SI/1e6
#            print param, truths_here[indx]

        # if 1d plots needed, make them
        if opts.plot_1d_extra:
            range_here = range_list[indx]
            # 1d PDF
            # Set range based on observed results in ALL sets of samples, by default
            fig =fig_1d_list[indx]
            ax = fig.gca()
            ax.set_xlabel(labels_tex[indx])
            ax.set_ylabel('$dP/d'+labels_tex[indx].replace('$','')+"$")
            try:
                my_kde = bounded_kde.BoundedKDE(dat_here,low=xmin,high=xmax)
                xvals = np.linspace(range_here[0],range_here[1],1000)
                yvals = my_kde.evaluate(xvals)
                ax.plot(xvals,yvals,color=my_cmap_values,linestyle= linestyle_list[pIndex])
                if opts.truth_file:
                    ax.axvline(truths_here[indx], color='k',linestyle='dashed')
            except:
                print " Failed to plot 1d KDE for ", labels_tex[indx]

            # 1d CDF
            fig =fig_1d_list_cum[indx]
            ax = fig.gca()
            ax.set_xlabel(labels_tex[indx])
            ax.set_ylabel('$P(<'+labels_tex[indx].replace('$','')+")$")
            xvals = dat_here
            #yvals = np.arange(len(dat_here))*1.0/len(dat_here)
            yvals = np.cumsum(weights)
            yvals = yvals/yvals[-1]
            ax.plot(xvals,yvals,color=my_cmap_values,linestyle= linestyle_list[pIndex] )
            if opts.truth_file:
                ax.axvline(truths_here[indx], color='k',linestyle='dashed')
            ax.set_xlim(xmin,xmax)

    # Add weight columns (unsorted) for overall unsorted plot
    weights = np.ones(len(dat_mass))*1.0/len(dat_mass)
    if 'weights' in samples.dtype.names:
        weights= samples['weights']
        weights = weights/np.sum(weights)
    # plot corner
#    smooth=smooth_list
    smooth1d=None
#    if opts.use_smooth_1d:
#        smooth1d=smooth_list
#        print smooth1d
    fig_base = corner.corner(dat_mass,smooth1d=smooth1d, range=range_list,weights=weights, labels=labels_tex, quantiles=quantiles_1d, plot_datapoints=False, plot_density=False, no_fill_contours=True, contours=True, levels=CIs,fig=fig_base,color=my_cmap_values ,hist_kwargs={'linestyle': linestyle_list[pIndex]}, linestyle=linestyle_list[pIndex],contour_kwargs={'linestyles':linestyle_list[pIndex]},truths=truths_here)


if opts.plot_1d_extra:
    for indx in np.arange(len(opts.parameter)):
        fig = fig_1d_list[indx]
        param = opts.parameter[indx]
        ax = fig.gca()
        # https://matplotlib.org/api/_as_gen/matplotlib.pyplot.legend.html
        if opts.use_legend:
            ax.legend(handles=line_handles, bbox_to_anchor=corner_legend_location, prop=corner_legend_prop,loc=2) 
        fig.savefig(param+fig_extension,dpi=dpi_base)

        fig = fig_1d_list_cum[indx]
        param = opts.parameter[indx]
        ax = fig.gca()
        # https://matplotlib.org/api/_as_gen/matplotlib.pyplot.legend.html
        if opts.use_legend:
            ax.legend(handles=line_handles, prop=corner_legend_prop, loc=4) # bbox_to_anchor=corner_legend_location, 
        fig.savefig(param+"_cum"+fig_extension,dpi=dpi_base)


if composite_list:
  for pIndex in [0]: # np.arange(len(composite_list)):  # should NEVER have more than one
    samples = composite_list[pIndex]
    # Create data for corner plot
    dat_mass = np.zeros( (len(samples["m1"]), len(labels_tex)) )
    lnL = samples["lnL"]
    indx_sorted = lnL.argsort()
    if len(lnL)<1:
        print " Failed to retrieve lnL for composite file ", composite_list[0]

    cm = plt.cm.get_cmap('rainbow') #'RdYlBu_r')
    y_span = lnL.max() - lnL.min()
    print " Composite file : lnL span ", y_span
    y_min = lnL.min()
    cm2 = lambda x: cm(  (x - y_min)/y_span)
    my_cmap_values = cm(    (lnL-y_min)/y_span) 
#    print my_cmap_values[:10]


    truths_here=None
    if opts.truth_file:
        truths_here =np.zeros(len(opts.parameter))
    for indx in np.arange(len(opts.parameter)):
        param = opts.parameter[indx]
        if param in field_names:
            dat_mass[:,indx] = samples[param]
        else:
            print " Trying alternative access for ", param
            dat_mass[:,indx] = extract_combination_from_LI(samples, param)
        # truths
        if opts.truth_file:
            param_to_extract = param
            if param in remap_LI_to_ILE.keys():
                param_to_extract  = remap_LI_to_ILE[param]
            truths_here[indx] = P_ref.extract_param(param_to_extract)
            if param in [ 'mc', 'm1', 'm2', 'mtotal']:
                truths_here[indx] = truths_here[indx]/lal.MSUN_SI
#            print param, truths_here[indx]

    print " Truths here ", truths_here
        

    # fix ranges
    if range_list == [] :
        range_list=None

    # reverse order ... make sure largest plotted last
    dat_mass = dat_mass[indx_sorted]   # Sort by lnL
    my_cmap_values = my_cmap_values[indx_sorted]
#    my_cmap_values = my_cmap_values[::-1]
#    dat_mass = dat_mass[::-1]
            
    # We will need to rewrite 'corner' to do what we want: see the source
    # https://github.com/dfm/corner.py/blob/master/corner/corner.py
    fig_base = our_corner.corner(dat_mass,range=range_list, plot_datapoints=True,weights=np.ones(len(dat_mass))*1.0/len(dat_mass), plot_density=False, no_fill_contours=True, plot_contours=False,contours=False,levels=None,fig=fig_base,data_kwargs={'color':my_cmap_values, 's':1}, truths=truths_here)

    # Create colorbar mappable
#    ax=plt.figure().gca()
#    ax.contourf(lnL, cm)

if opts.use_legend and opts.posterior_label:
    plt.legend(handles=line_handles, bbox_to_anchor=corner_legend_location, prop=corner_legend_prop,loc=4)
#plt.colorbar()  # Will fail, because colors not applied

# title
if opts.use_title:
    print " Addding title ", opts.use_title
    plt.title(opts.use_title)

param_postfix = "_".join(opts.parameter)
res_base = len(opts.parameter)*dpi_base
matplotlib.rcParams.update({'font.size': 11+int(len(opts.parameter)), 'legend.fontsize': legend_font_base+int(1.3*len(opts.parameter))})   # increase font size if I have more panels, to keep similar aspect
plt.savefig("corner_"+param_postfix+fig_extension,dpi=res_base)        # use more resolution, to make sure each image remains of consistent quality



from __future__ import division
import cupy
import os

#ILE_base = os.environ["ILE_CODE_PATH"]  # default: store code inside main repo. Maintainer controls.
_cuda_code = None


def Q_inner_product_cupy(Q, A, start_indices, window_size):
    num_time_points, num_lms = Q.shape
    num_extrinsic_samples, _ = A.shape

    assert not cupy.isfortran(Q)
    assert not cupy.isfortran(A)

    out = cupy.empty(
        (num_extrinsic_samples, window_size),
        dtype=cupy.complex128,
        order="C",
    )

    global _cuda_code
    if _cuda_code is None:
        # it's assumed that cuda_Q_inner_product.cu is placed in the same folder as this code
        path = os.path.join(os.path.dirname(__file__), 'cuda_Q_inner_product.cu')
        # alternative to deal with packaging in another directory
        if not (os.path.isfile(path)):
            path = os.path.join(os.path.split(os.path.dirname(__file__))[0], 'cuda_Q_inner_product.cu')
        with open(path, 'r') as f:
            _cuda_code = f.read()
            Q_prod_fn = cupy.RawKernel(_cuda_code, "Q_inner")
    else:
        Q_prod_fn = cupy.RawKernel(_cuda_code, "Q_inner")

    float_prec = 16
    num_threads_x = 4
    num_threads_y = 1024 // 4
    block_size = num_threads_x, num_threads_y, 0
    grid_size = (
        (num_extrinsic_samples+num_threads_x-1)//num_threads_x,
        0,
        0,
    )
    args = (
        Q, A, start_indices, window_size,
        num_time_points, num_extrinsic_samples, num_lms,
        out,
    )
    Q_prod_fn(
        grid_size, block_size, args,
        shared_mem=cupy.int32(num_threads_x*num_lms*float_prec),
    )

    return out



#! /usr/bin/env python
#
# Do not use this code.

import sys

import numpy as np

from glue.ligolw import utils, lsctables, table, ligolw
lsctables.use_in(ligolw.LIGOLWContentHandler)
from lal.series import read_psd_xmldoc

import lal

import RIFT.lalsimutils as lalsimutils
import RIFT.misc.ourparams as ourparams
opts,  rosDebugMessagesDictionary = ourparams.ParseStandardArguments()

import common_cl

def estimate_adaptive_exponent(rho2Net, nskip):
    """
     Estimate the 'beta' parameter needed to regularize the likelihood, so even the max-likelihood event doesn't overwhelm the histogram Add an ad-hoc x2 to be in better agreement with manual calibration for zero-noise MDC event

    exp(rho^2/2)^beta ~ nchunk*10
    """
    return min(0.8, 2*np.log(nskip*10)*2./rho2Net)

if __file__ == sys.argv[0]:

    rho2Net = 0

    if opts.coinc:
        # Extract trigger SNRs
        ci_table = table.get_table(utils.load_filename(opts.coinc,contenthandler =lalsimutils.cthdler), lsctables.CoincInspiralTable.tableName)
        rho2Net = ci_table[0].snr**2

    if opts.inj and opts.channel_name:
        # Make fake signal.  Use one instrument.
    
        # Read injection parameters
        sim_inspiral_table = table.get_table(utils.load_filename(opts.inj,contenthandler =lalsimutils.cthdler), lsctables.SimInspiralTable.tableName)
        Psig = lalsimutils.ChooseWaveformParams()
        Psig.copy_lsctables_sim_inspiral(sim_inspiral_table[opts.event_id])
    
        data_fake_dict ={}
        psd_dict = {}
        rhoFake = {}
        fminSNR =opts.fmin_SNR
        fmaxSNR =opts.fmax_SNR
        fSample = opts.srate
        fNyq = fSample/2.
        rho2Net =0
        # Insure signal duration and hence frequency spacing specified
        Psig.deltaF = lalsimutils.findDeltaF(Psig)
    
        if opts.psd_file is not None:
            xmldoc = utils.load_filename(opts.psd_file)
            psd_dict = read_psd_xmldoc(xmldoc)

            # Remove unwanted PSDs
            if opts.channel_name is not None:
                dets = common_cl.parse_cl_key_value(opts.channel_name).keys()
                for det in psd_dict.keys():
                    if det not in dets:
                        del psd_dict[det]
                    
        else:
            psd_dict = common_cl.parse_cl_key_value(opts.psd_file_singleifo)
            for det, psdf in psd_dict.iteritems():
                psd_dict[det] = lalsimutils.get_psd_series_from_xmldoc(psdf, det)

        """
        # FIXME: Reenable with ability to parse out individual PSD functions from
        # lalsim --- steal from ligolw_inj_snr
        if not(opts.psd_file) and not(opts.psd_file_singleifo):
            analyticPSD_Q = True # For simplicity, using an analytic PSD
            psd_dict[det] = lalsim.SimNoisePSDiLIGOSRD   
        """

        for det in psd_dict:
            Psig.detector = det
            data_fake_dict[det] = lalsimutils.non_herm_hoff(Psig)

            deltaF = data_fake_dict[det].deltaF
            fmin = psd_dict[det].f0
            fmax = fmin + psd_dict[det].deltaF*len(psd_dict[det].data)-deltaF
            psd_dict[det] = lalsimutils.resample_psd_series(psd_dict[det], deltaF)
        
            rhoFake[det] = lalsimutils.singleIFOSNR(data_fake_dict[det], psd_dict[det], fSample/2, fminSNR, fmaxSNR)

            # FIXME: Reenable with options parsing
            #if opts.verbose:
                #print det, rhoFake[det]

            # Construct rho^2
            rho2Net += rhoFake[det]*rhoFake[det]

     # FIXME: Reenable with options parsing
    #if opts.verbose:
        #print "network, squared: %f" % rho2Net
    print estimate_adaptive_exponent(rho2Net, opts.nskip)
